from fastapi import (
    Depends,
    HTTPException,
    Response,
    status,
)

from tuspyserver.file import TusUploadFile


def termination_extension_routes(router, options):
    """
    https://tus.io/protocols/resumable-upload#termination
    """

    @router.delete("/{uuid}", status_code=status.HTTP_204_NO_CONTENT)
    def extension_termination_route(
        uuid: str, response: Response, _=Depends(options.auth)
    ) -> Response:
        file = TusUploadFile(uid=uuid, options=options)

        # Check if the upload ID is valid
        if not file.exists:
            raise HTTPException(status_code=404, detail="Upload not found")

        # Delete the file and metadata for the upload from the mapping
        file.delete()

        # Return a 204 No Content response
        response.headers["Tus-Resumable"] = options.tus_version
        response.status_code = status.HTTP_204_NO_CONTENT

        return response

    return router



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    from tuspyserver.file import TusUploadFile

from tuspyserver.params import TusUploadParams

import os
import json


class TusUploadInfo:
    _params: TusUploadParams | None
    file: TusUploadFile

    def __init__(self, file: TusUploadFile, params: TusUploadParams | None = None):
        self.file = file
        self._params = params
        # create if doesn't exist
        if params and not self.exists:
            self.serialize()

    @property
    def params(self):
        self.deserialize()
        return self._params

    @params.setter
    def params(self, value):
        self._params = value
        self.serialize()

    @property
    def path(self) -> str:
        return os.path.join(self.file.options.files_dir, f"{self.file.uid}.info")

    @property
    def exists(self) -> bool:
        return os.path.exists(self.path)

    def serialize(self) -> None:
        with open(self.path, "w") as f:
            json_string = json.dumps(
                self._params, indent=4, default=lambda k: k.__dict__
            )
            f.write(json_string)

    def deserialize(self) -> TusUploadParams | None:
        if self.exists:
            with open(self.path, "r") as f:
                json_dict = json.load(f) or {}
                self._params = TusUploadParams(**json_dict)

        return None



from typing import Hashable
from pydantic import BaseModel


class TusUploadParams(BaseModel):
    metadata: dict[Hashable, str]
    size: int | None
    offset: int = 0
    upload_part: int = 0
    created_at: str
    defer_length: bool = False
    upload_chunk_size: int = 0
    expires: float | str | None



from typing import Callable, Optional

from fastapi import APIRouter
from pydantic import BaseModel

from tuspyserver.routes.core import core_routes
from tuspyserver.routes.creation import creation_extension_routes
from tuspyserver.routes.termination import termination_extension_routes


class TusRouterOptions(BaseModel):
    prefix: str
    files_dir: str
    max_size: int
    auth: Callable[[], None] | None
    days_to_keep: int
    on_upload_complete: Callable[[str, dict], None] | None
    upload_complete_dep: Callable[..., Callable[[str, dict], None]] | None
    tags: list[str] | None
    tus_version: str
    tus_extension: str


async def noop():
    pass


def create_tus_router(
    prefix: str = "files",
    files_dir="/tmp/files",
    max_size=128849018880,
    auth: Optional[Callable[[], None]] = noop,
    days_to_keep: int = 5,
    on_upload_complete: Optional[Callable[[str, dict], None]] = None,
    upload_complete_dep: Optional[Callable[..., Callable[[str, dict], None]]] = None,
    tags: Optional[list[str]] = None,
):
    async def _fallback_on_complete_dep() -> Callable[[str, dict], None]:
        return on_upload_complete or (lambda *_: None)

    upload_complete_dep = upload_complete_dep or _fallback_on_complete_dep

    options = TusRouterOptions(
        prefix=prefix[1:] if prefix and prefix[0] == "/" else prefix,
        files_dir=files_dir,
        max_size=max_size,
        auth=auth,
        days_to_keep=days_to_keep,
        on_upload_complete=on_upload_complete,
        upload_complete_dep=upload_complete_dep
        or (lambda _: on_upload_complete or (lambda *_: None)),
        tags=tags,
        tus_version="1.0.0",
        tus_extension=",".join(
            [
                "creation",
                "creation-defer-length",
                "creation-with-upload",
                "expiration",
                "termination",
            ]
        ),
    )

    clean_prefix = prefix.lstrip("/").rstrip("/")
    router = APIRouter(
        prefix=clean_prefix,
        redirect_slashes=True,
        tags=options.tags or ["Tus"],
    )

    modules = [
        core_routes,
        # extensions
        creation_extension_routes,
        termination_extension_routes,
    ]

    for mod in modules:
        router = mod(router, options)

    return router



import base64
import inspect
import os
from datetime import datetime, timedelta
from typing import Callable


from fastapi import (
    Depends,
    Header,
    HTTPException,
    Response,
    status,
)

from tuspyserver.file import TusUploadFile
from tuspyserver.request import make_request_chunks_dep


def core_routes(router, options):
    """
    https://tus.io/protocols/resumable-upload#core-protocol
    """

    request_chunks_dep = make_request_chunks_dep(options)

    @router.head("/{uuid}", status_code=status.HTTP_200_OK)
    def core_head_route(
        response: Response, uuid: str, _=Depends(options.auth)
    ) -> Response:
        # validate file

        file = TusUploadFile(uid=uuid, options=options)
        if file.info is None or not file.exists:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

        # encode metadata

        filename = file.info.metadata.get("filename") or file.info.metadata.get("name")
        if filename is None:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="Upload-file.metadata missing required field: filename",
            )

        filetype = file.info.metadata.get("filetype") or file.info.metadata.get("type")
        if filetype is None:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="Upload-Metadata missing required field: filetype",
            )

        def b64(s: str) -> str:
            return base64.b64encode(s.encode("utf-8")).decode("ascii")

        # construct response

        response.headers["Tus-Resumable"] = file.options.tus_version
        response.headers["Upload-Metadata"] = (
            f"filename {b64(filename)}, filetype {b64(filetype)}"
        )
        response.headers["Upload-Length"] = str(file.info.size)
        response.headers["Upload-Offset"] = str(file.info.offset)
        response.headers["Content-Length"] = str(file.info.size)
        response.headers["Cache-Control"] = "no-store"

        response.status_code = status.HTTP_200_OK

        return response

    @router.patch("/{uuid}", status_code=status.HTTP_204_NO_CONTENT)
    async def core_patch_route(
        response: Response,
        uuid: str,
        content_length: int = Header(None),
        upload_offset: int = Header(None),
        _=Depends(request_chunks_dep),
        __=Depends(options.auth),
        on_complete: Callable[[str, dict], None] = Depends(options.upload_complete_dep),
    ) -> Response:
        file = TusUploadFile(uid=uuid, options=options)

        # check if the upload ID is valid
        if not file.info or uuid != file.uid:
            raise HTTPException(status_code=404)

        # check if the Upload Offset with Content-Length header is correct
        if file.info.offset != upload_offset + content_length:
            raise HTTPException(status_code=409)

        # init copy of params to update
        new_params = file.info

        if file.info.defer_length:
            new_params.size = upload_offset

        if not file.info.expires:
            date_expiry = datetime.now() + timedelta(days=options.days_to_keep)
            new_params.expires = str(date_expiry.isoformat())

        # save param changes
        file.info = new_params

        if file.info.size == file.info.offset:
            response.headers["Tus-Resumable"] = options.tus_version
            response.headers["Upload-Offset"] = str(
                str(file.info.offset) if file.info.offset > 0 else str(content_length)
            )
            response.headers["Upload-Expires"] = str(file.info.expires)
            response.status_code = status.HTTP_204_NO_CONTENT
            if options.on_upload_complete:
                options.on_upload_complete(
                    os.path.join(options.files_dir, f"{uuid}"),
                    file.info.metadata,
                )
        else:
            response.headers["Tus-Resumable"] = options.tus_version
            response.headers["Upload-Offset"] = str(file.info.offset)
            response.headers["Upload-Expires"] = str(file.info.expires)
            response.status_code = status.HTTP_204_NO_CONTENT

        if file.info and file.info.size == file.info.offset:
            file_path = os.path.join(options.files_dir, uuid)
            if options.on_upload_complete is None:
                result = on_complete(file_path, file.info.metadata)
                # if the callback returned a coroutine, await it
                if inspect.isawaitable(result):
                    await result

        return response

    @router.options("/", status_code=status.HTTP_204_NO_CONTENT)
    def core_options_route(response: Response, __=Depends(options.auth)) -> Response:
        # create response headers
        response.headers["Tus-Version"] = options.tus_version
        response.headers["Tus-Resumable"] = options.tus_version
        response.headers["Tus-Extension"] = options.tus_extension
        response.headers["Tus-Max-Size"] = str(options.max_size)
        response.headers["Content-Length"] = str(0)
        response.status_code = status.HTTP_204_NO_CONTENT

        return response

    return router



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    from tuspyserver.router import TusRouterOptions


from fastapi import (
    HTTPException,
    Path,
    Request,
)

from starlette.requests import ClientDisconnect

from tuspyserver.file import TusUploadFile


def make_request_chunks_dep(options: TusRouterOptions):
    async def request_chunks_dep(
        request: Request,
        uuid: str = Path(...),
        post_request: bool = False,
    ) -> bool | None:
        # init file handle
        file = TusUploadFile(uid=uuid, options=options)

        # check if valid file
        if not file.exists or not file.info:
            raise HTTPException(status_code=404, detail="Upload not found")

        # init variables
        has_chunks = False
        new_params = file.info

        # process chunk stream
        with open(f"{options.files_dir}/{uuid}", "ab") as f:
            try:
                async for chunk in request.stream():
                    has_chunks = True
                    # skip empty chunks but continue processing
                    if len(chunk) == 0:
                        continue
                    # Check if upload would exceed declared size
                    if (
                        new_params.size is not None
                        and new_params.offset + len(chunk) > new_params.size
                    ):
                        raise HTTPException(
                            status_code=400,
                            detail="Upload would exceed declared Upload-Length",
                        )
                    # throw if max size exceeded
                    if new_params.offset + len(chunk) > options.max_size:
                        raise HTTPException(
                            status_code=413,
                            detail="Upload exceeds maximum allowed size",
                        )
                    # write chunk otherwise
                    f.write(chunk)
                    # update upload params
                    new_params.offset += len(chunk)
                    new_params.upload_chunk_size = len(chunk)
                    new_params.upload_part += 1
                    # save updated params
                    file.info = new_params
            except ClientDisconnect:
                return False
            except Exception as e:
                # save the error
                new_params.error = str(e)
                # save updated params
                file.info = new_params

                return False
            finally:
                f.close()

        # For empty files in a POST request, we still want to return True
        # to ensure the file gets created properly
        if post_request and not has_chunks:
            # update params for empty file
            new_params.offset = 0
            new_params.upload_chunk_size = 0
            new_params.upload_part += 1
            # save updated params
            file.info = new_params

        return True

    return request_chunks_dep


def get_request_headers(request: Request, uuid: str) -> dict:
    proto = "http"
    host = request.headers.get("host")
    if request.headers.get("X-Forwarded-Proto") is not None:
        proto = request.headers.get("X-Forwarded-Proto")
    if request.headers.get("X-Forwarded-Host") is not None:
        host = request.headers.get("X-Forwarded-Host")

    # Get the prefix from the request path
    path_parts = request.url.path.split("/")
    prefix = path_parts[1] if len(path_parts) > 1 else "files"

    return {
        "location": f"{proto}://{host}/{prefix}/{uuid}",
        "proto": proto,
        "host": host,
    }



import base64
import inspect
import os
from datetime import datetime, timedelta
from typing import Callable

from fastapi import (
    Depends,
    Header,
    HTTPException,
    Request,
    Response,
    status,
)

from tuspyserver.file import TusUploadFile, TusUploadParams
from tuspyserver.request import get_request_headers


def creation_extension_routes(router, options):
    """
    https://tus.io/protocols/resumable-upload#creation
    """

    @router.post("", status_code=status.HTTP_201_CREATED)
    @router.post("/", status_code=status.HTTP_201_CREATED)
    async def extension_creation_route(
        request: Request,
        response: Response,
        upload_metadata: str = Header(None),
        upload_length: int = Header(None),
        upload_defer_length: int = Header(None),
        _=Depends(options.auth),
        on_complete: Callable[[str, dict], None] = Depends(options.upload_complete_dep),
    ) -> Response:
        # validate upload defer length
        if upload_defer_length is not None and upload_defer_length != 1:
            raise HTTPException(status_code=400, detail="Invalid Upload-Defer-Length")
        # set expiry date
        date_expiry = datetime.now() + timedelta(days=options.days_to_keep)
        # create upload metadata
        metadata = {}
        if upload_metadata is not None and upload_metadata != "":
            # Decode the base64-encoded string
            for kv in upload_metadata.split(","):
                key, value = kv.rsplit(" ", 1)
                decoded_value = base64.b64decode(value.strip()).decode("utf-8")
                metadata[key.strip()] = decoded_value
        # create upload params
        params = TusUploadParams(
            metadata=metadata,
            size=upload_length,
            offset=0,
            upload_part=0,
            created_at=str(datetime.now()),
            defer_length=upload_defer_length is not None,
            expires=str(date_expiry.isoformat()),
        )
        # create the file
        file = TusUploadFile(options=options, params=params)
        # update request headers
        response.headers["Location"] = get_request_headers(
            request=request, uuid=file.uid
        )["location"]
        response.headers["Tus-Resumable"] = options.tus_version
        response.headers["Content-Length"] = str(0)
        # set status code
        response.status_code = status.HTTP_201_CREATED
        # run completion hooks
        if file.info and file.info.size == 0:
            file_path = os.path.join(options.files_dir, file.uid)
            result = on_complete(file_path, file.info.metadata)
            # if the callback returned a coroutine, await it
            if inspect.isawaitable(result):
                await result

        return response

    return router



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    from tuspyserver.router import TusRouterOptions

import os
import datetime

from uuid import uuid4

from tuspyserver.params import TusUploadParams
from tuspyserver.info import TusUploadInfo


class TusUploadFile:
    uid: str
    _info: TusUploadInfo
    _options: TusRouterOptions

    def __init__(
        self,
        options: TusRouterOptions,
        uid: str | None = None,
        params: TusUploadParams | None = None,
    ):
        self._options = options
        # init
        if uid is None:
            # creating new file
            self.uid = str(uuid4().hex)
            self.create()
        else:
            # reading existing file
            self.uid = uid
        # create the files dir if necessary
        if not os.path.exists(self._options.files_dir):
            os.makedirs(self._options.files_dir)
        # instantiate upload info
        self._info = TusUploadInfo(file=self, params=params)

    @property
    def path(self) -> str:
        return os.path.join(self._options.files_dir, f"{self.uid}")

    @property
    def options(self) -> TusRouterOptions:
        return self._options

    @property
    def info(self) -> TusUploadParams:
        return self._info.params

    @info.setter
    def info(self, value) -> None:
        self._info.params = value

    @property
    def exists(self) -> bool:
        return os.path.exists(self.path)

    def create(self) -> None:
        open(self.path, "a").close()

    def read(self) -> bytes | None:
        if self.exists:
            with open(self.path, "rb") as f:
                return f.read()
        return None

    def delete(self, uid: str) -> None:
        if os.path.exists(self.path):
            os.remove(self.path)

        if os.path.exists(self.info.path):
            os.remove(self.info.path)

    def __len__(self) -> int:
        if self.exists:
            return os.path.getsize(self.path)
        return 0


def list_files(options: TusRouterOptions) -> list[str]:
    return [f for f in os.listdir(options.files_dir) if len(f) == 32]


def gc_files(options: TusRouterOptions):
    for uid in list_files():
        file = TusUploadFile(uid=uid, options=options)
        if (
            file.info
            and file.info.expires
            and datetime.datetime.fromisoformat(file.info.expires)
            < datetime.datetime.now()
        ):
            file.delete(uid)
